import sys


def removeprefix(s: str, prefix: str) -> str:
    if sys.version_info >= (3, 9):  # pragma: py-lt-39
        return s.removeprefix(prefix)
    else:  # pragma: py-gte-39
        if s.startswith(prefix):
            return s[len(prefix) :]
        return s



from __future__ import annotations

import base64
import dataclasses
import json
import urllib.parse
from typing import Any, Dict

import datalad_cds.compat


@dataclasses.dataclass
class Spec:
    dataset: str
    sub_selection: dict

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Spec:
        return cls(dataset=d["dataset"], sub_selection=d["sub-selection"])

    @classmethod
    def from_json(cls, s: str) -> Spec:
        return cls.from_dict(json.loads(s))

    @classmethod
    def from_url(cls, url: str) -> Spec:
        if not url.startswith("cds:v1-"):
            raise ValueError("unsupported URL value encountered")
        spec = cls.from_json(
            base64.urlsafe_b64decode(
                urllib.parse.unquote(
                    datalad_cds.compat.removeprefix(url, "cds:v1-")
                ).encode("utf-8")
            ).decode("utf-8")
        )
        return spec

    def to_dict(self) -> Dict[str, Any]:
        return {
            "dataset": self.dataset,
            "sub-selection": self.sub_selection,
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), separators=(",", ":"), sort_keys=True)

    def to_url(self) -> str:
        json_spec = self.to_json()
        url = "cds:v1-" + urllib.parse.quote(
            base64.urlsafe_b64encode(json_spec.encode("utf-8"))
        )
        return url



import pathlib

import datalad.api as da
import pytest
from _pytest.fixtures import SubRequest
from datalad.conftest import setup_package  # noqa: F401


@pytest.fixture
def empty_dataset(tmp_path: pathlib.Path) -> da.Dataset:
    dataset = da.create(tmp_path)
    dataset.configuration("set", [("remote.cds.dry-run", "true")], scope="local")
    yield dataset
    dataset.drop(what="all", reckless="kill", recursive=True)


@pytest.fixture(params=["non-lazy", "lazy"])
def single_file_dataset(request: SubRequest, empty_dataset: da.Dataset) -> da.Dataset:
    request_dict = {
        "dataset": "reanalysis-era5-pressure-levels",
        "sub-selection": {
            "variable": "temperature",
            "pressure_level": "1000",
            "product_type": "reanalysis",
            "date": "2017-12-01/2017-12-31",
            "time": "12:00",
            "format": "grib",
        },
    }
    dataset = empty_dataset
    dataset.download_cds(
        request_dict,
        lazy=request.param == "lazy",
        path="download.grib",
    )
    if request.param == "lazy":
        dataset.configuration(
            "set",
            [("remote.cds.annex-security-allow-unverified-downloads", "ACKTHPPT")],
            scope="local",
        )
    return dataset



import subprocess

import datalad.api as da


def test_with_git_annex_testremote(single_file_dataset: da.Dataset) -> None:
    dataset = single_file_dataset
    dataset.get("download.grib")
    result = subprocess.run(
        [
            "git",
            "-C",
            dataset.path,
            "annex",
            "testremote",
            "cds",
            "--test-readonly",
            "download.grib",
        ],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0, f"{result.stdout=}\n{result.stderr=}"



import pathlib
import subprocess

import cdsapi
from annexremote import Master, ProtocolError, RemoteError, SpecialRemote

import datalad_cds.spec

CDS_REMOTE_UUID = "923e2755-e747-42f4-890a-9c921068fb82"


class CDSRemote(SpecialRemote):
    transfer_store = None
    remove = None

    def initremote(self) -> None:
        pass

    def prepare(self) -> None:
        pass

    def _is_dry_run(self) -> bool:
        try:
            remote_name = self.annex.getgitremotename()
        except ProtocolError:
            return False
        git_dir = self.annex.getgitdir()
        result = subprocess.run(
            [
                "git",
                "--git-dir={}".format(git_dir),
                "config",
                "--get",
                "remote.{}.dry-run".format(remote_name),
            ],
            capture_output=True,
            text=True,
        )
        return result.returncode == 0 and result.stdout.strip() == "true"

    def _retrieve_cds(self, spec: datalad_cds.spec.Spec, filename: str) -> None:
        if self._is_dry_run():
            pathlib.Path(filename).write_text(spec.to_json())
            return
        c = cdsapi.Client()
        c.retrieve(spec.dataset, spec.sub_selection, filename)

    def transfer_retrieve(self, key: str, filename: str) -> None:
        urls = self.annex.geturls(key, "cds:")
        exceptions = []
        for url in urls:
            try:
                self._retrieve_cds(datalad_cds.spec.Spec.from_url(url), filename)
                break
            except Exception as e:
                exceptions.append(e)
        else:
            raise RemoteError(exceptions)

    def whereis(self, key: str) -> str:
        url = self.annex.geturls(key, "cds:")[0]
        return datalad_cds.spec.Spec.from_url(url).to_json()

    def checkpresent(self, key: str) -> bool:
        urls = self.annex.geturls(key, "cds:")
        return len(urls) > 0

    def claimurl(self, url: str) -> bool:
        return url.startswith("cds:")

    def checkurl(self, url: str) -> bool:
        return url.startswith("cds:")

    def getcost(self) -> int:
        # This is a very expensive remote
        return 1000

    def getavailability(self) -> str:
        # The Climate Data Store is publicly available on the internet
        return "global"


def main() -> None:
    master = Master()
    remote = CDSRemote(master)
    master.LinkRemote(remote)
    master.Listen()



"""DataLad CDS extension"""

__docformat__ = "restructuredtext"

import logging

lgr = logging.getLogger("datalad.cds")

# Defines a datalad command suite.
# This variable must be bound as a setuptools entrypoint
# to be found by datalad
command_suite = (
    # description of the command suite, displayed in cmdline help
    "Custom Downloader for CDS data-store",
    [
        # specification of a command, any number of commands can be defined
        (
            # importable module that contains the command implementation
            "datalad_cds.download_cds",
            # name of the command class implementation in above module
            "DownloadCDS",
            # optional name of the command in the cmdline API
            "download-cds",
            # optional name of the command in the Python API
            "download_cds",
        ),
    ],
)

from . import _version  # noqa: E402

__version__ = _version.get_versions()["version"]



# -*- coding: utf-8 -*-
#
# datalad_cds documentation build configuration file, created by
# sphinx-quickstart on Tue Oct 13 08:41:19 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import subprocess
import sys
from os import pardir
from os.path import abspath, dirname, exists
from os.path import join as opj
from typing import List

import datalad_cds

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# generate missing pieces
for setup_py_path in (
    opj(pardir, "setup.py"),  # travis
    opj(pardir, pardir, "setup.py"),
):  # RTD
    if exists(setup_py_path):
        sys.path.insert(0, abspath(dirname(setup_py_path)))
        # Build manpage
        try:
            subprocess.run(
                args=[
                    setup_py_path,
                    "build_manpage",
                    "--cmdsuite",
                    "datalad_cds:command_suite",
                    "--manpath",
                    abspath(opj(dirname(setup_py_path), "build", "man")),
                    "--rstpath",
                    opj(dirname(__file__), "generated", "man"),
                ],
                check=True,
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            # shut up and do your best
            pass

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]

# for the module reference
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "DataLad CDS extension"
copyright = "2018-{}, Matthias Riße, the DataLad team and contributors".format(
    datetime.datetime.now().year
)
author = "Matthias Riße"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
version = datalad_cds.__version__
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns: List[str] = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/datalad_logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If true, the index is split into individual pages for each letter.
html_split_index = True

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# smart quotes are incompatible with the RST flavor of the generated manpages
# but see `smartquotes_action` for more fine-grained control, in case
# some of this functionality is needed
smartquotes = False



import hypothesis as h
import pytest

from datalad_cds.spec import Spec


@h.given(...)
def test_spec_url_equality(spec: Spec) -> None:
    assert Spec.from_url(spec.to_url()) == spec


@h.given(...)
def test_spec_invalid_url_causes_value_error(url: str) -> None:
    h.assume(not url.startswith("cds:v1-"))
    with pytest.raises(ValueError):
        Spec.from_url(url)



import json
import os
from typing import Union

import datalad.api as da
import pytest

import datalad_cds

request_dict = {
    "dataset": "reanalysis-era5-pressure-levels",
    "sub-selection": {
        "variable": "temperature",
        "pressure_level": "1000",
        "product_type": "reanalysis",
        "date": "2017-12-01/2017-12-31",
        "time": "12:00",
        "format": "grib",
    },
}


@pytest.mark.parametrize("cds_request", [request_dict, json.dumps(request_dict)])
def test_download_cds(cds_request: Union[str, dict], empty_dataset: da.Dataset) -> None:
    dataset = empty_dataset
    dataset.download_cds(
        cds_request,
        path="download.grib",
    )
    actual_request = datalad_cds.spec.Spec.from_json(
        (dataset.pathobj / "download.grib").read_text()
    )
    if isinstance(cds_request, dict):
        expected_request = datalad_cds.spec.Spec.from_dict(cds_request)
    elif isinstance(cds_request, str):
        expected_request = datalad_cds.spec.Spec.from_json(cds_request)
    assert actual_request == expected_request


@pytest.mark.parametrize("cds_request", [request_dict, json.dumps(request_dict)])
def test_download_cds_lazy(
    cds_request: Union[str, dict], empty_dataset: da.Dataset
) -> None:
    dataset = empty_dataset
    dataset.download_cds(
        cds_request,
        path="download.grib",
        lazy=True,
    )
    assert (
        os.readlink(dataset.pathobj / "download.grib")
        == ".git/annex/objects/5J/pV/URL--cds&cv1-eyJkYXRhc2V0IjoicmVhbmFs-77566133ebfe9220aefbeed5a58b6972/URL--cds&cv1-eyJkYXRhc2V0IjoicmVhbmFs-77566133ebfe9220aefbeed5a58b6972"
    )



#!/usr/bin/env python

import sys

sys.path.insert(0, "")

from setuptools import setup  # noqa: E402

import versioneer  # noqa: E402
from _datalad_buildsupport.setup import BuildManPage  # noqa: E402

cmdclass = versioneer.get_cmdclass()
cmdclass.update(build_manpage=BuildManPage)

if __name__ == "__main__":
    setup(
        name="datalad-cds",
        version=versioneer.get_version(),
        cmdclass=cmdclass,
    )



"""DataLad extension for the Climate Data Store"""

__docformat__ = "restructuredtext"
import logging
from typing import Iterable, Literal, Optional, Union

from datalad.distribution.dataset import (
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.interface.base import Interface, build_doc, eval_results
from datalad.interface.common_opts import nosave_opt, save_message_opt
from datalad.interface.results import get_status_dict
from datalad.support.annexrepo import AnnexRepo
from datalad.support.constraints import EnsureNone, EnsureStr
from datalad.support.param import Parameter

import datalad_cds.cds_remote
import datalad_cds.spec

logger = logging.getLogger("datalad.cds.download_cds")


# decoration auto-generates standard help
@build_doc
# all commands must be derived from Interface
class DownloadCDS(Interface):
    """Downloads specified datasets from the CDS data store"""

    _params_ = dict(
        spec=Parameter(
            doc="""A json string or python dictionary containing the key
            "dataset" with the datasets name (i.e. what is shown as the first
            parameter to cdsapi.Client.retrieve if you do a "Show API request"
            on some dataset in the CDS) and the key "sub-selection" with the
            sub-selection of the dataset that should be fetched (i.e. what is
            shown as the second parameter to cdsapi.Client.retrieve).""",
        ),
        dataset=Parameter(
            args=("-d", "--dataset"),
            metavar="PATH",
            doc="""specify the dataset to add files to. If no dataset is given,
            an attempt is made to identify the dataset based on the current
            working directory. Use [CMD: --nosave CMD][PY: save=False PY] to
            prevent adding files to the dataset.""",
            constraints=EnsureDataset() | EnsureNone(),
        ),
        path=Parameter(
            args=("-O", "--path"),
            doc="""target path to download to.""",
            constraints=EnsureStr(),
        ),
        lazy=Parameter(
            args=("--lazy",),
            action="store_true",
            doc="""By default the file will be immediately downloaded. If the
            lazy flag is supplied then the CDS request is only recorded as a
            source for the file, but no download is initiated. Keep in mind that
            there is no way to validate the correctness of the request if the
            lazy flag is used.""",
        ),
        save=nosave_opt,
        message=save_message_opt,
    )

    @staticmethod
    @datasetmethod(name="download_cds")
    @eval_results
    def __call__(
        spec: Union[str, dict],
        path: str,
        *,
        dataset: Optional[str] = None,
        message: Optional[str] = None,
        save: bool = True,
        lazy: bool = False,
    ) -> Iterable[dict]:
        if isinstance(spec, dict):
            parsed_spec = datalad_cds.spec.Spec.from_dict(spec)
        elif isinstance(spec, str):
            parsed_spec = datalad_cds.spec.Spec.from_json(spec)
        else:
            raise TypeError("spec could not be parsed")
        ds = require_dataset(dataset, check_installed=True)
        ensure_special_remote_exists_and_is_enabled(ds.repo, "cds")
        pathobj = ds.pathobj / path
        url = parsed_spec.to_url()
        options = []
        if lazy:
            options.append("--fast")
        ds.repo.add_url_to_file(pathobj, url, options=options)
        if save:
            msg = (
                message
                if message is not None
                else "[DATALAD] Download from Climate Data Store"
            )
            yield ds.save(pathobj, message=msg)
        yield get_status_dict(action="cds", ds=ds, status="ok")


def ensure_special_remote_exists_and_is_enabled(
    repo: AnnexRepo, remote: Literal["cds"]
) -> None:
    """Initialize and enable the cds special remote, if it isn't already.

    Very similar to datalad.customremotes.base.ensure_datalad_remote.
    """

    uuids = {"cds": datalad_cds.cds_remote.CDS_REMOTE_UUID}
    uuid = uuids[remote]

    name = repo.get_special_remotes().get(uuid, {}).get("name")
    if not name:
        repo.init_remote(
            remote,
            [
                "encryption=none",
                "type=external",
                "autoenable=true",
                "externaltype={}".format(remote),
                "uuid={}".format(uuid),
            ],
        )
    elif repo.is_special_annex_remote(name, check_if_known=False):
        logger.debug("special remote %s is enabled", name)
    else:
        logger.debug("special remote %s found, enabling", name)
        repo.enable_remote(name)
